@@ -16,15 +16,17 @@ from natsort import natsorted
 
 from koyo.typing import PathLike
 
+ALLOW_EXTRA_ARGS = {"help_option_names": ["-h", "--help"], "ignore_unknown_options": True, "allow_extra_args": True}
 
-def expand_data_dirs(input_dir: str) -> ty.List[str]:
+
+def expand_data_dirs(input_dir: str) -> list[str]:
     """Expand data directory."""
     if "*" in input_dir:
-        return glob.glob(input_dir)
+        return list(glob.glob(input_dir))
     return [input_dir]
 
 
-def parse_paths(input_dirs: list[PathLike], absolute: bool = False, sort: bool = False) -> ty.List[str]:
+def parse_paths(input_dirs: list[PathLike], absolute: bool = False, sort: bool = False) -> list[str]:
     """Parse list/tuple of paths and check whether any of them have glob-like pattern."""
     result = []
     for path in input_dirs:
@@ -42,16 +44,24 @@ def parse_paths(input_dirs: list[PathLike], absolute: bool = False, sort: bool =
     return result
 
 
-def cli_parse_paths(ctx, param, value) -> ty.List[str]:
+def cli_parse_paths(ctx, param, value) -> list[str]:
     """Parse paths."""
     return parse_paths(value)
 
 
-def cli_parse_paths_sort(ctx, param, value) -> ty.List[str]:
+def cli_parse_paths_sort(ctx, param, value) -> list[str]:
     """Parse paths."""
     return parse_paths(value, sort=True)
 
 
+def cli_parse_paths_sort_auto_glob(ctx, param, value) -> list[str]:
+    """Parse paths."""
+    if len(value) == 1:
+        if Path(value[0]).is_dir():
+            value = [f"{value[0]}/*"]
+    return parse_paths(value, sort=True)
+
+
 def with_plugins(plugins, **kwargs):
     """
     A decorator to register external CLI commands to an instance of
@@ -100,16 +110,15 @@ class BrokenCommand(click.Command):
         """Define the special help messages after instantiating a `click.Command()`."""
         click.Command.__init__(self, name)
 
-        util_name = os.path.basename(sys.argv and sys.argv[0] or __file__)
+        util_name = os.path.basename((sys.argv and sys.argv[0]) or __file__)
 
         if os.environ.get("CLICK_PLUGINS_HONESTLY"):  # pragma no cover
-            icon = "\U0001F4A9"
+            icon = "\U0001f4a9"
         else:
             icon = "\u2020"
 
         self.help = (
-            "\nWarning: entry point could not be loaded. Contact "
-            "its author for help.\n\n\b\n" + traceback.format_exc()
+            "\nWarning: entry point could not be loaded. Contact its author for help.\n\n\b\n" + traceback.format_exc()
         )
         self.short_help = icon + f" Warning: could not load plugin. See `{util_name} {self.name} --help`."
 
@@ -137,13 +146,28 @@ _verbosity_options = [
 ]
 
 
-def verbosity_options(func):
+def verbosity_options(func: ty.Callable) -> ty.Callable:
     """Verbosity options."""
     for option in reversed(_verbosity_options):
         func = option(func)
     return func
 
 
+def dev_options(func: ty.Callable) -> ty.Callable:
+    """Setup dev options."""
+    from koyo.system import IS_PYINSTALLER
+
+    if not IS_PYINSTALLER:
+        func = click.option(
+            "--dev",
+            help="Flat to indicate that CLI should run in development mode and catch all errors.",
+            default=False,
+            is_flag=True,
+            show_default=True,
+        )(func)
+    return func
+
+
 def repr_filelist(filelist: ty.Sequence[PathLike]) -> str:
     """Pretty iterable..."""
     res = [Path(file).name for file in filelist]
@@ -161,7 +185,7 @@ def repr_iterable(iterable: ty.Sequence[ty.Any]):
         return iterable
 
 
-def append(table: ty.List[ty.Tuple[str, str, str]], name: str = "", param: str = "", value: ty.Any = "") -> None:
+def append(table: list[tuple[str, str, str]], name: str = "", param: str = "", value: ty.Any = "") -> None:
     """Append to table."""
     if isinstance(value, Path):
         value = str(value)
@@ -176,7 +200,7 @@ def append(table: ty.List[ty.Tuple[str, str, str]], name: str = "", param: str =
     table.append((name, param, value))
 
 
-def format_value(description: str, args: str, value: ty.Any) -> ty.List[ty.Tuple[str, str, str]]:
+def format_value(description: str, args: str, value: ty.Any) -> list[tuple[str, str, str]]:
     """Format value."""
     res = []
     # lists should be printed as multiple rows
@@ -225,20 +249,18 @@ def get_args_from_option(option: ty.Callable) -> str:
 class Parameter:
     """Parameter object."""
 
-    __slots__ = ["description", "args", "value"]
+    __slots__ = ["args", "description", "value"]
 
-    def __init__(self, description: str, args: ty.Union[str, ty.Callable] | None, value: ty.Optional[ty.Any] = None):
+    def __init__(self, description: str, args: str | ty.Callable | None, value: ty.Any | None = None):
         self.description = description
         self.args = get_args_from_option(args) if callable(args) else args
         self.value = value
 
-    def with_value(
-        self, value: ty.Any, description: ty.Optional[str] = None, args: ty.Optional[str] = None
-    ) -> Parameter:
+    def with_value(self, value: ty.Any, description: str | None = None, args: str | None = None) -> Parameter:
         """Set the value of the parameter and return self."""
         return Parameter(description or self.description, args or self.args, value)
 
-    def to_list(self) -> ty.List[ty.Tuple[str, str, str]]:
+    def to_list(self) -> list[tuple[str, str, str]]:
         """Return formatted version of the value."""
         if self.value is not None:
             return format_value(self.description, self.args, self.value)
@@ -272,6 +294,11 @@ def print_parameters(*parameters: Parameter, log: bool = True, silent: bool = Fa
             logger.info(line)
 
 
+def exception_msg(msg: str):
+    """Display error message."""
+    logger.exception(msg)
+
+
 def error_msg(msg: str):
     """Display error message."""
     logger.error(msg)
@@ -304,11 +331,11 @@ def expand_dirs(input_dir: str) -> ty.Sequence[str]:
 
 
 # noinspection PyUnusedLocal
-def arg_parse_path(ctx, param, value) -> ty.List[Path]:
+def arg_parse_path(ctx, param, value) -> list[Path]:
     """Split arguments."""
     if value is None:
         return []
-    res: ty.List[str] = []
+    res: list[str] = []
     if isinstance(value, (str, Path)):
         value = [value]
     for path in value:
@@ -346,7 +373,15 @@ def arg_split_int(ctx, param, value):
     return args
 
 
-def parse_str_framelist(framelist: str) -> ty.List[int]:
+# noinspection PyUnusedLocal
+def arg_parse_env(ctx, param, value) -> tuple[str | None, ty.Any | None]:
+    """Split argument into environment variables."""
+    if value is None:
+        return None, None
+    return [v.split("=", maxsplit=1) for v in value]
+
+
+def parse_str_framelist(framelist: str) -> list[int]:
     """Parse list provided in string format."""
     if framelist is None:
         return []
@@ -376,7 +411,7 @@ def arg_parse_framelist(ctx, param, value: str):
     return parse_str_framelist(value)
 
 
-def arg_parse_framelist_multi(ctx, param, value: ty.Tuple[str]):
+def arg_parse_framelist_multi(ctx, param, value: tuple[str]):
     """Parse framelist."""
     if value is None:
         return None
@@ -422,12 +457,12 @@ def timed_iterator(
             func(f"[{i}/{n_tasks}] {_text} {format_human_time_s(execution_time)} [{avg}{rem}{tot}]")
 
 
-def parse_arg(arg: str, key: str):
+def parse_arg(arg: str, key: str) -> tuple[str, ty.Any]:
     """Parse argument."""
     try:
         if key:
             arg = arg.split(key)[1]
-        name, value = arg.split("=")
+        name, value = arg.split("=", maxsplit=1)
         # try parsing value - it will fail if string value was provided
         try:
             value = literal_eval(value)
@@ -442,9 +477,9 @@ def parse_arg(arg: str, key: str):
         raise ValueError(f"Could not parse argument {arg}")
 
 
-def parse_extra_args(extra_args: ty.Optional[ty.Tuple[str]]):
+def parse_extra_args(extra_args: tuple[str, ...] | None) -> dict[str, ty.Any]:
     """Arguments"""
-    kwargs = {}
+    kwargs: dict[str, ty.Any] = {}
     if extra_args is None:
         return kwargs
     for arg in extra_args:
@@ -452,12 +487,23 @@ def parse_extra_args(extra_args: ty.Optional[ty.Tuple[str]]):
             info_msg(f"Skipping argument {arg} as it does not contain '='")
             continue
         name, value = parse_arg(arg, "")
-        kwargs[name] = value
+        if name in kwargs:
+            if name in kwargs and not isinstance(kwargs[name], list):
+                kwargs[name] = [kwargs[name]]
+            if isinstance(kwargs[name], list):
+                kwargs[name].append(value)
+            elif isinstance(kwargs[name], (str, int, float, bool)):
+                kwargs[name] = value
+            else:
+                kwargs[name] = [kwargs[name], value]
+        else:
+            kwargs[name] = value
     return kwargs
 
 
-def parse_fig_args(extra_args: ty.Optional[ty.Tuple[str]], clean: bool = False):
-    """Parse extra parameters."""
+def parse_args_with_keys(
+    extra_args: tuple[str] | None, fields: tuple[str, ...], clean: bool = False
+) -> dict[str, ty.Any] | tuple[dict[str, ty.Any], tuple[str]]:
     extra_kwargs = {}
     if extra_args is None:
         if clean:
@@ -465,12 +511,12 @@ def parse_fig_args(extra_args: ty.Optional[ty.Tuple[str]], clean: bool = False):
         return extra_kwargs
     extra_args_ = []
     for arg in extra_args:
-        if arg.startswith("--fig:"):
-            name, value = parse_arg(arg, "--fig:")
-            extra_kwargs[name] = value
-        elif arg.startswith("--f:"):
-            name, value = parse_arg(arg, "--f:")
-            extra_kwargs[name] = value
+        if any(arg.startswith(k) for k in fields):
+            for k in fields:
+                if arg.startswith(k):
+                    name, value = parse_arg(arg, k)
+                    extra_kwargs[name] = value
+                    break
         else:
             extra_args_.append(arg)
     if clean:
@@ -478,45 +524,79 @@ def parse_fig_args(extra_args: ty.Optional[ty.Tuple[str]], clean: bool = False):
     return extra_kwargs
 
 
-def parse_env_args(extra_args: ty.Optional[ty.Tuple[str]], clean: bool = False):
+def parse_fig_args(
+    extra_args: tuple[str] | None, clean: bool = False
+) -> dict[str, ty.Any] | tuple[dict[str, ty.Any], tuple[str]]:
+    """Parse extra parameters."""
+    return parse_args_with_keys(extra_args, ("--fig:", "--f:"), clean)
+
+
+@ty.overload
+def parse_env_args(extra_args: tuple[str, ...] | None, clean: bool = False) -> dict[str, ty.Any]:
+    """Parse extra environment variables."""
+    ...
+
+
+@ty.overload
+def parse_env_args(extra_args: tuple[str, ...] | None, clean: bool = False) -> tuple[dict[str, ty.Any], tuple[str]]:
+    """Parse extra environment variables."""
+    ...
+
+
+def parse_env_args(
+    extra_args: tuple[str, ...] | None, clean: bool = False
+) -> dict[str, ty.Any] | tuple[dict[str, ty.Any], tuple[str]]:
     """Parse extra environment variables."""
     env_kwargs = {}
     if extra_args is None:
         if clean:
             return env_kwargs, extra_args
         return env_kwargs
-    extra_args_ = []
+
+    parsed_extra_args = []
     for arg in extra_args:
         if arg.startswith("--env:"):
             name, value = parse_arg(arg, "--env:")
             env_kwargs[name] = value
         else:
-            extra_args_.append(arg)
+            parsed_extra_args.append(arg)
     if clean:
-        return env_kwargs, extra_args_
+        return env_kwargs, parsed_extra_args
     return env_kwargs
 
 
-def set_env_args(**kwargs):
+def set_env_args(**kwargs: ty.Any) -> None:
     """Set environment variables."""
-    from loguru import logger
-
     for name, value in kwargs.items():
         os.environ[name] = str(value)
-        logger.trace(f"Set environment variable: {name}={value}")
+        logger.info(f"Set environment variable: {name}={value}")
 
 
-def exit_with_error(skip_error: bool = False):
+def set_env_args_from_tuples(env_vars: tuple[str | None, str | None]) -> None:
+    """Set environment variables."""
+    for name, value in env_vars:
+        if name is None or value is None:
+            continue
+        os.environ[name] = str(value)
+        logger.info(f"Set environment variable: {name}={value}")
+
+
+def filter_kwargs(*allowed: str, **kwargs: ty.Any) -> dict[str, ty.Any]:
+    """Filter allowed kwargs."""
+    return {key: value for key, value in kwargs.items() if key in allowed}
+
+
+def exit_with_error(skip_error: bool = False) -> int | None:
     """Skip error or exit with error."""
     import sys
 
     if skip_error:
-        return
+        return sys.exit()
     return sys.exit(1)
 
 
 def select_from_list(
-    item_list: ty.List[ty.Any],
+    item_list: list[ty.Any],
     text: str = "Please select one of the options from the list",
     auto_select: str = "off",
     default: int = -1,
@@ -535,3 +615,15 @@ def select_from_list(
         elif auto_select.lower() == "oldest":
             return 0
     return default
+
+
+# Commonly used click options
+extra_args_ = click.argument("extra_args", nargs=-1, type=click.UNPROCESSED)
+env_var_ = click.option(
+    "-e",
+    "--env",
+    type=str,
+    multiple=True,
+    help="Environment variables to set. Values should be set as KEY=VALUE (e.g. --env KEY=VALUE).",
+    callback=arg_parse_env,
+)
