@@ -2,10 +2,22 @@
 
 import gc
 import os
+import pathlib
 import sys
 from contextlib import contextmanager, redirect_stdout
 
 
+@contextmanager
+def fix_pathlib():
+    """Fix pathlib on Windows."""
+    posix_backup = pathlib.PosixPath
+    try:
+        pathlib.PosixPath = pathlib.WindowsPath
+        yield
+    finally:
+        pathlib.PosixPath = posix_backup
+
+
 @contextmanager
 def hide_stdout():
     """Hide stdout."""
@@ -27,6 +39,12 @@ def nullout():
         sys.stderr = save_stderr
 
 
+@contextmanager
+def noop(*args, **kwargs):
+    """Function."""
+    yield
+
+
 @contextmanager
 def no_gc():
     """Context manager to disable garbage collection."""
