@@ -1,18 +1,118 @@
 """Generic utilities."""
+
 from __future__ import annotations
 
 import re
 import typing as ty
-import unicodedata
 from collections.abc import Iterable
-from difflib import get_close_matches
 from math import ceil, floor
+from pathlib import Path
+from random import shuffle
 
 import numba as nb
 import numpy as np
 from natsort import natsorted
+from loguru import logger
 
-from koyo.typing import SimpleArrayLike
+from koyo.typing import PathLike, SimpleArrayLike
+from koyo.system import running_as_pyinstaller_app
+
+if ty.TYPE_CHECKING:
+    from koyo.fig_mixin import PptxPdfWrapper
+
+
+def human_readable_byte_size(n_bytes: int) -> str:
+    """
+
+    Parameters
+    ----------
+    n_bytes : int
+        Number of bytes.
+
+    Returns
+    -------
+    Human readable string : str
+
+    """
+    suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]
+    i = 0
+    while n_bytes >= 1024 and i < len(suffixes) - 1:
+        n_bytes /= 1024.0
+        i += 1
+    f = (f"{n_bytes:.2f}").rstrip("0").rstrip(".")
+    return f"{f} {suffixes[i]}"
+
+
+def split_tooltip(text: str, max_width: int = 80) -> str:
+    """Split tooltip text into multiple lines based on maximum width.
+
+    Args:
+        text: Text to be split
+        max_width: Maximum number of characters per line (default: 80)
+
+    Returns:
+        Formatted string with line breaks
+    """
+    if not text:
+        return ""
+
+    words = text.split()
+    lines = []
+    current_line = []
+    current_width = 0
+    for word in words:
+        word_len = len(word)
+        if current_width + word_len + 1 <= max_width:  # +1 for space
+            current_line.append(word)
+            current_width += word_len + 1
+        else:
+            if current_line:
+                lines.append(" ".join(current_line))
+            current_line = [word]
+            current_width = word_len + 1
+
+    if current_line:
+        lines.append(" ".join(current_line))
+    return "\n".join(lines)
+
+
+def get_unique_without_sort(seq: list) -> list:
+    """Get unique values without sorting."""
+    seen = set()
+    seen_add = seen.add
+    return [x for x in seq if not (x in seen or seen_add(x))]
+
+
+def ensure_output_dir_exists(
+    output_dir: PathLike | None, pptx_or_pdf: PptxPdfWrapper | None = None
+) -> tuple[bool, Path]:
+    """Ensure that the output directory exists.
+
+    Returns
+    -------
+    bool
+        whether the directory was created
+    Path
+        output directory
+    """
+    # check if the output directory is actually needed - not if exporting as PDF or PPTX
+    if pptx_or_pdf and pptx_or_pdf.as_pptx_or_pdf:
+        return False, Path("")
+    if output_dir:
+        output_dir = Path(output_dir)
+        output_dir.mkdir(parents=True, exist_ok=True)
+        return True, output_dir
+    return False, Path("")
+
+
+def clean_path(path: str) -> Path:
+    """Clean path that might have been stored in a wrong way."""
+    path = str(path)
+    # replace Windows backslashes with forward slashes
+    path = path.replace("\\", "/")
+    # replace MacOS network path with local path and resolve the network drive if it exists
+    path = path.replace("/Volumes/", "/")
+    return Path(path)
 
 
 def is_installed(module: str) -> bool:
@@ -26,11 +126,48 @@ def is_installed(module: str) -> bool:
     return loader is not None
 
 
+def get_version(module: str) -> str:
+    """Get current version of package."""
+    import importlib.metadata
+
+    try:
+        installed_version = importlib.metadata.version(module)
+    except importlib.metadata.PackageNotFoundError:
+        return "N/A"
+    return installed_version
+
+
+def is_above_version(module: str, version: str) -> bool:
+    """Check whether the module is above a certain version."""
+    import importlib.metadata
+
+    from packaging.version import Version
+
+    try:
+        installed_version = importlib.metadata.version(module)
+    except importlib.metadata.PackageNotFoundError:
+        logger.warning(f"Module {module} not found.")
+        return False
+    installed_version = Version(installed_version)
+    version = Version(version)
+    if installed_version is not None:
+        return installed_version >= version
+    logger.warning(f"Module {module} not found.")
+    return False
+
+
 def get_format(fmt: str) -> str:
     """Parse format."""
     return fmt if fmt.startswith(".") else f".{fmt}"
 
 
+def ensure_list(array: np.ndarray | list[np.ndarray]) -> list[np.ndarray]:
+    """Ensure that the input is a list."""
+    if isinstance(array, np.ndarray):
+        return [array]
+    return array
+
+
 def find_nearest_divisor(
     value: int | float,
     divisor: int | float = 1,
@@ -69,12 +206,12 @@ def find_nearest_divisor(
     return -1
 
 
-def slugify_name(value: str):
+def slugify_name(value: str) -> str:
     """Slugify filename."""
     return value.replace(" ", "_").replace("/", "_").replace("\\", "_").replace(":", "-")
 
 
-def order_parameters(**kwargs):
+def order_parameters(**kwargs: ty.Any) -> dict:
     """Order parameters."""
     kwargs_ = {}
     for key in natsorted(kwargs):
@@ -82,7 +219,7 @@ def order_parameters(**kwargs):
     return kwargs_
 
 
-def exclude_parameters(exclude: ty.Iterable[str], **kwargs):
+def exclude_parameters(exclude: ty.Iterable[str], **kwargs: ty.Any) -> dict:
     """Exclude parameters."""
     kwargs_ = {}
     for key in natsorted(kwargs):
@@ -102,7 +239,7 @@ def get_module_path(module: str, filename: str) -> str:
     return path
 
 
-def reraise_exception_if_debug(exc, message: str, env_key: str = "DEV_MODE"):
+def reraise_exception_if_debug(exc, message: str = "Exception occurred", env_key: str = "DEV_MODE") -> None:
     """Reraise exception if debug mode is enabled and jump into the debugger."""
     import os
 
@@ -144,7 +281,7 @@ def flatten_nested_list(list_of_lists: list[list]) -> list:
     return [item for sublist in list_of_lists for item in sublist]
 
 
-def get_list_difference(li1: list, li2: list):
+def get_list_difference(li1: list, li2: list) -> list:
     """Get difference between two lists."""
     # get difference between two lists while keeping the order
     li_dif = [i for i in li1 + li2 if i not in li1 or i not in li2]
@@ -218,16 +355,23 @@ def find_nearest_value_single(data: SimpleArrayLike, value: int | float) -> int
     return data[idx]
 
 
-def find_nearest_index_batch(array: SimpleArrayLike, values: SimpleArrayLike) -> np.ndarray:
+def find_nearest_value_in_dict(data: dict[float, ty.Any], value: int | float) -> ty.Any:
+    """Find nearest value in dictionary."""
+    key = find_nearest_value(list(data.keys()), value)
+    return data[key]
+
+
+def find_nearest_index_batch(array: SimpleArrayLike, values: SimpleArrayLike, sort: bool = False) -> np.ndarray:
     """Find nearest index."""
     # make sure array is a numpy array
     array = np.asarray(array)
     values = np.asarray(values)
     if not array.size or not values.size:
-        return np.array([])
+        return np.array([], dtype=np.int64)
 
     # get insert positions
-    indices = np.searchsorted(array, values, side="left")
+    sorter = np.argsort(array) if sort else None
+    indices = np.searchsorted(array, values, side="left", sorter=sorter)
 
     # find indexes where previous index is closer
     prev_idx_is_less = (indices == len(array)) | (
@@ -235,6 +379,8 @@ def find_nearest_index_batch(array: SimpleArrayLike, values: SimpleArrayLike) ->
         < np.fabs(values - array[np.minimum(indices, len(array) - 1)])
     )
     indices[prev_idx_is_less] -= 1
+    if sorter is not None:
+        indices = sorter[indices]
     return indices
 
 
@@ -245,7 +391,21 @@ def find_nearest_value(data: ty.Iterable, value: int | (float | (np.ndarray | It
     return data[idx]
 
 
-def get_kws(func: ty.Callable, **kwargs) -> dict:
+def get_pixels_within_radius(grid: np.ndarray, x: int, y: int, radius: int = 0) -> np.ndarray:
+    """Get pixels within radius."""
+    # Get the bounds for the sub-grid (handling edge cases)
+    y_min = max(0, y - radius)
+    y_max = min(grid.shape[1], y + radius + 1)
+    x_min = max(0, x - radius)
+    x_max = min(grid.shape[0], x + radius + 1)
+
+    # Extract the sub-grid around the point (y, x)
+    sub_grid = grid[y_min:y_max, x_min:x_max].flatten()
+    # drop NaNs, Infs, etc
+    return sub_grid[np.isfinite(sub_grid)]
+
+
+def get_kws(func: ty.Callable, **kwargs: ty.Any) -> dict:
     """Get kwargs."""
     import inspect
 
@@ -258,21 +418,21 @@ def get_kws(func: ty.Callable, **kwargs) -> dict:
     return kws
 
 
-def format_count(value):
+def format_count(value: float) -> str:
     """Format count."""
     if value < 1e3:
         return f"{value:.0f}"
     elif value < 1e6:
-        return f"{value/1e3:.1f}K"
+        return f"{value / 1e3:.1f}K"
     elif value < 1e9:
-        return f"{value/1e6:.1f}M"
-    return f"{value/1e9:.1f}B"
+        return f"{value / 1e6:.1f}M"
+    return f"{value / 1e9:.1f}B"
 
 
 def format_size(size: int) -> str:
     """Convert bytes to nicer format."""
     if size < 2**10:
-        return "%s" % size
+        return f"{size}"
     elif size < 2**20:
         return "%.1fK" % (size / float(2**10))
     elif size < 2**30:
@@ -284,7 +444,7 @@ def format_size(size: int) -> str:
     return "%.1fP" % (size / float(2**50))
 
 
-def is_number(value):
+def is_number(value: ty.Any) -> bool:
     """Quick and easy way to check if input is a number.
 
     Parameters
@@ -300,7 +460,7 @@ def is_number(value):
     return isinstance(value, (int, float, complex))
 
 
-def check_value_order(value_min, value_max):
+def check_value_order(value_min: float, value_max: float) -> tuple[float, float]:
     """Check whether the value order is correct (min -> max).
 
     Parameters
@@ -325,7 +485,7 @@ def check_value_order(value_min, value_max):
     return value_min, value_max
 
 
-def get_value(new_value, current_value):
+def get_value(new_value: ty.Any, current_value: ty.Any) -> ty.Any:
     """Get value."""
     if new_value is None:
         return current_value
@@ -411,7 +571,7 @@ def rescale_value(
     return ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min
 
 
-def chunks(item_list, n_items: int = 0, n_tasks: int = 0):
+def chunks(item_list: ty.Iterable[ty.Any], n_items: int = 0, n_tasks: int = 0) -> ty.Generator:
     """Yield successive n-sized chunks from `item_list`.
 
     Parameters
@@ -436,9 +596,26 @@ def chunks(item_list, n_items: int = 0, n_tasks: int = 0):
         yield item_list[i : i + n_items]
 
 
-def zip_chunks(item_list, n_items: int, *items):
+def random_chunks(item_list: ty.Iterable[ty.Any], n_items: int = 0, n_tasks: int = 0) -> ty.Generator:
+    """Randomly yield successive n-sized chunks from `item_list`.
+
+    Warning: This function will shuffle the list ahead of time.
+    """
+    if n_items == 0 and n_tasks == 0:
+        raise ValueError("You must specify either 'n_items' or 'n_tasks'. Both are zero")
+    if n_tasks:
+        n_items = ceil(len(item_list) / n_tasks)
+    shuffle(item_list)
+    for i in range(0, len(item_list), n_items):
+        yield item_list[i : i + n_items]
+
+
+def zip_chunks(item_list, n_items: int = 0, n_tasks: int = 0, *items):
     """Yield successive n-sized chunks from `item_list`."""
-    yield from zip(chunks(item_list, n_items), *[chunks(x, n_items) for x in items])
+    yield from zip(
+        chunks(item_list, n_items=n_items, n_tasks=n_tasks),
+        *[chunks(x, n_items=n_items, n_tasks=n_tasks) for x in items],
+    )
 
 
 def sequential_chunks(item_list, n_items: int):
@@ -466,20 +643,20 @@ def get_min_max(values: np.ndarray | ty.Iterable | list) -> tuple[int | float, i
     return np.min(values), np.max(values)
 
 
-def need_rotation(array: np.ndarray) -> bool:
+def need_rotation(array: np.ndarray | tuple[int, ...]) -> bool:
     """Check whether image needs to be rotated."""
-    shape = array.shape
+    shape = array.shape if not isinstance(array, tuple) else array
     if len(shape) == 3:
         return shape[1] > shape[2]
     return shape[0] > shape[1]
 
 
-def rotate(array: np.ndarray, auto_rotate: bool):
+def rotate(array: np.ndarray, auto_rotate: bool) -> np.ndarray:
     """Rotate but only if user requested rotation."""
     return array if not auto_rotate else check_image_orientation(array)
 
 
-def check_image_orientation(array):
+def check_image_orientation(array: np.ndarray) -> np.ndarray:
     """Transpose image if the primary size is larger than the secondary size in order to improve images.
 
     Parameters
@@ -499,12 +676,14 @@ def check_image_orientation(array):
     # return np.rot90(array) if shape[0] > shape[1] else array
 
 
-def slugify(value, allow_unicode=False):
+def slugify(value: ty.Any, allow_unicode: bool = False) -> str:
     """Convert to ASCII if 'allow_unicode' is False.
 
     Convert spaces or repeated dashes to single dashes. Remove characters that aren't alphanumerics, underscores,
     or hyphens. Convert to lowercase. Also strip leading and trailing whitespace, dashes, and underscores.
     """
+    import unicodedata
+
     value = str(value)
     if allow_unicode:
         value = unicodedata.normalize("NFKC", value)
@@ -519,7 +698,42 @@ def get_array_mask(array: np.ndarray, min_val: float, max_val: float):
     return np.logical_and(array >= min_val, array <= max_val)
 
 
-def get_array_window(array: np.ndarray, min_val: float, max_val: float, *arrays):
+@nb.njit(parallel=True)
+def get_array_mask_(array: np.ndarray, min_val: int | float, max_val: int | float) -> np.ndarray:
+    """Return mask for array."""
+    mask = np.empty(array.shape, dtype=np.bool_)
+    for i in nb.prange(array.size):
+        val = array[i]
+        mask[i] = min_val <= val <= max_val
+    return mask
+
+
+@nb.njit(parallel=True)
+def filter_array(
+    array: np.ndarray, array_to_mask: np.ndarray, min_val: int | float, max_val: int | float
+) -> np.ndarray:
+    """Filter array based on another array."""
+    # First, count how many elements pass the condition
+    count = 0
+    for i in nb.prange(array_to_mask.size):
+        val = array_to_mask[i]
+        if min_val <= val <= max_val:
+            count += 1
+
+    # Allocate an output array of the required size
+    result = np.zeros(count, dtype=array.dtype)
+
+    # Second pass: fill the result with filtered values
+    idx = 0
+    for i in nb.prange(array_to_mask.size):
+        val = array_to_mask[i]
+        if min_val <= val <= max_val:
+            result[idx] = array[i]
+            idx += 1
+    return result
+
+
+def get_array_window(array: np.ndarray, min_val: float, max_val: float, *arrays: np.ndarray) -> list[np.ndarray]:
     """Get narrower view of array based on upper and lower limits.
 
     The first array is the one that is used to create mask.
@@ -608,21 +822,24 @@ def difference_matrix(a: np.ndarray) -> np.ndarray:
     return x - x.transpose()
 
 
-def running_as_pyinstaller_app() -> bool:
-    """Infer whether we are running pyinstaller bundle."""
-    import sys
 
-    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")
-
-
-def get_close_matches_case(word, possibilities, *args, **kwargs):
+def get_close_matches_case(word, possibilities, *args: ty.Any, **kwargs: ty.Any):
     """Case-insensitive version of difflib.get_close_matches."""
+    from difflib import get_close_matches
+
     lower_word = word.lower()
     lower_possibilities = {p.lower(): p for p in possibilities}
     lower_matches = get_close_matches(lower_word, lower_possibilities.keys(), *args, **kwargs)
     return [lower_possibilities[m] for m in lower_matches]
 
 
+def get_close_matches_case_msg(word: str, possibilities: ty.Iterable[str]) -> str:
+    """Get message for close matches."""
+    match = get_close_matches_case(word, possibilities)
+    match = f"Did you mean: `{', '.join(match)}`?" if match else ""
+    return f"Unknown parameter: `{word}`.{match}"
+
+
 def calculate_array_size(a, as_str: bool = True):
     """Calculate the size of an array in Mb."""
     if not hasattr(a, "nbytes"):
@@ -667,6 +884,7 @@ def get_distributed_list(
     framelist: np.ndarray | None = None,
     as_index: bool = False,
     trim: bool = False,
+    offset: int = 0,
 ) -> np.ndarray:
     """Get list of frames that span the entire dataset.
 
@@ -682,6 +900,8 @@ def get_distributed_list(
         if `True`, the distributed framelist will consist of indices rather than actual frame IDs
     trim : bool, optional
         if `True`, the distributed framelist will be trimmed to the maximum available frame
+    offset : int
+        Offset  to apply to the start of the frame list.
 
     Returns
     -------
@@ -689,6 +909,7 @@ def get_distributed_list(
         array with list of frames to extract from the dataset
     """
     start = 0 if as_index else 1
+    start += offset
     end = total
 
     # check whether n_frames is not a fraction
