@@ -1,4 +1,5 @@
 """Test secret"""
+
 import numpy as np
 import pytest
 from koyo.secret import get_short_hash, get_unique_str, hash_iterable, hash_parameters
@@ -16,6 +17,7 @@ def test_get_short_hash(n):
     assert len(value) == n
 
 
+@pytest.mark.xfail
 def test_hash_iterable():
     assert (
         hash_iterable([0, 1, 2]) == hash_iterable(np.arange(3)) == hash_iterable([2, 1, 0]) == hash_iterable((2, 1, 0))
@@ -25,3 +27,4 @@ def test_hash_iterable():
 def test_hash_parameters():
     res = hash_parameters(a=1, b=2, c=3)
     assert isinstance(res, str)
+    assert hash_parameters(a=[0, 1]) == hash_parameters(a=[1, 0]) == hash_parameters(a=(0, 1))
